/*
 * Copyright 2001  Brian S. Dean <bsd@bsdhome.com>
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY BRIAN S. DEAN ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL BRIAN S. DEAN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 */

/*----------------------------------------------------------------------
  $Id: dbregc.c,v 1.1.1.1 2001/06/24 17:48:23 bsd Exp $
  ----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  dbregc.c

  Author : Brian S. Dean
  Date   : 24 June, 2001

  IA32 Debug register calculator.  Specify breakpoint type, address,
  etc on the command line and dbreg prints out the appropriate dr0-dr7
  values.

  Usage: dbregc [options]

  Where options can be:

    -h                     : print usage information

    -7 dr7                 : specify existing dr7 value

    -b R:AA:L:0xADDRESS:BB : specify a watchpoint

         R       = specify breakpoint register number (0 - 3)
         AA      = "rw", "wo", or "ex"
         L       = watchpoint length (1, 2, or 4)
         ADDRESS = hexadecimal address of the watchpoint

    -w 0xADDRESS,LEN       : specify a watchpoint (alternate format);
                             a write-only watchpoint is assumed.

         ADDRESS = hexadecimal address of the watchpoint
         LEN     = length of the watch area (1 - 16)

  The -b option may be specified multiple times

  ----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <limits.h>
#include <machine/reg.h>
#include <machine/sysarch.h>


char * progname;


struct watch_t {
  int     active; /* 1 or 0 (true or false) */
  int     len;    /* 1, 2, or 4 */
  int     type;   /* See DBREG_DR7_* from <machine/reg.h> */
  caddr_t addr;   /* address of watch point */
};


char * watchtype_str(int type)
{
  switch ( type ) {
    case DBREG_DR7_EXEC   : return "execute   "; break;
    case DBREG_DR7_RDWR   : return "read/write"; break;
    case DBREG_DR7_WRONLY : return "write     "; break;
    default               : return "invalid   "; break;
  }
}


int usage(void)
{
  fprintf(stderr, 
          "\n"
          "Usage: %s [options]\n"
          "\n"
          "  Where options may be:\n"
          "\n"
          "     -h                 : print usage information\n"
          "\n"
          "     -7  dr7            : specify existing dr7 value\n"
          "\n"
          "    -b R:AA:L:0xADDRESS : specify a watchpoint\n"
          "\n"
          "         R       = specify breakpoint register number (0 - 3)\n"
          "         AA      = \"rw\", \"wo\", or \"ex\"\n"
          "         L       = watchpoint length (1, 2, or 4)\n"
          "         ADDRESS = hexadecimal address of the watchpoint\n"
          "\n"
          "    -w 0xADDRESS,LEN    : specify watchpoint (alternate format);\n"
          "                          a write-only watcpoint is assumed\n"
          "\n"
          "         ADDRESS = hexadecimal address of the watchpoint\n"
          "         LEN     = length of the watch area (1 - 16)\n"
          "\n"
          "  The -b option can be specified multiple times\n"
          "\n",
          progname);
  return 0;
}



int parse_watch(struct watch_t * w, int nw, char * wspec)
{
  int wnum;
  char acc[3];
  int len;
  u_int32_t addr;
  int n;

  /* parse: 0:rw:4:0x00000000 */
  /* parse: 0:wo:4:0x00000000 */
  /* parse: 0:ex:4:0x00000000 */

  n = sscanf(wspec, "%d:%c%c:%d:0x%x", &wnum, &acc[0], &acc[1], &len, &addr);
  if (n != 5) {
    fprintf(stderr, "%s: can't scan breakpoint spec \"%s\"\n", 
            progname, wspec);
    return -1;
  }

  acc[2] = 0;

  if (wnum >= 4) {
    return -2;
  }

  switch ( len ) {
    case 1:
    case 2:
    case 4:
      break;
    default:
      fprintf(stderr, "%s: invalid watchpoint length = %d\n", progname, len);
      return -3;
      break;
  }

  if (strcmp(acc,"rw") == 0) {
    w[wnum].type = DBREG_DR7_RDWR;
  }
  else if (strcmp(acc,"wo") == 0) {
    w[wnum].type = DBREG_DR7_WRONLY;
  }
  else if (strcmp(acc,"ex") == 0) {
    w[wnum].type = DBREG_DR7_EXEC;
  }
  else {
    fprintf(stderr, "%s: invalid watchpoint type = \"%s\"\n", progname, acc);
    return -4;
  }

  w[wnum].active  = 1;
  w[wnum].len     = len;
  w[wnum].addr    = (caddr_t)addr;

  return 0;
}


int parse_watch_alt(struct watch_t * w, int nw, char * wspec)
{
  int len, save_len;
  u_int32_t addr;
  int n;
  int i;
  int size;

  /* parse: 0x00000000,26 */

  n = sscanf(wspec, "0x%x,%d", &addr, &save_len);
  if (n != 2) {
    return -1;
  }

  len = save_len;

  for (i=0; i<nw && len; i++) {
    if (!w[i].active) {
      size = 4;
      if (len < size)
        size = len;
      w[i].active = 1;
      w[i].type = DBREG_DR7_WRONLY;
      w[i].addr = (caddr_t)addr;
      if (size == 3)
        w[i].len  = 4;
      else
        w[i].len  = size;
      addr += size;
      len  -= size;
    }
  }

  if (len) {
    fprintf(stderr, "%s: size specification too large\n", progname);
    fprintf(stderr, "     not enough registers to watch %d bytes; "
            "you can watch %d bytes\n",
            save_len, save_len-len);
    return -1;
  }


  return 0;
}



int set_dbregs(unsigned int dr7, struct watch_t * w, int nw, struct dbreg * d)
{
  int i;
  int rc;

  memset(d, 0, sizeof(*d));

  d->dr7 = dr7;

  for (i=0; i<nw; i++) {
    if (w[i].active) {
      rc = i386_set_watch(i, (unsigned int)w[i].addr, w[i].len, 
                          w[i].type, d);
      if (rc < 0) {
        fprintf(stderr,
                "%s: can't set up debug registers for watch point %d\n",
                progname, i);
      }
    }
  }

  return 0;
}


int main ( int argc, char * argv[] )
{
  int rc, i, ch, errors;
  unsigned int dr7;
  char * p;
  struct dbreg d;
  struct watch_t w[4];

  memset(&w, 0, sizeof(w));
  dr7 = 0;

  progname = rindex(argv[0],'/');
  if (progname) {
    progname++;
  }
  else {
    progname = argv[0];
  }

  if (argc < 2) {
    usage();
    exit(1);
  }

  errors = 0;
  while ((ch = getopt(argc,argv,"7:b:hw:")) != -1) {
    switch (ch) {
      case '7':
        dr7 = strtoul(optarg, &p, 0);
        if (*p != 0) {
          fprintf(stderr, "%s: can't scan dr7 value from \"%s\"\n",
                  progname, optarg);
          exit(1);
        }
        break;
      case 'b':
        rc = parse_watch ( w, 4, optarg );
        if (rc < 0) {
          errors = 1;
        }
        break;
      case 'w':
        rc = parse_watch_alt ( w, 4, optarg );
        if (rc < 0) {
          errors = 1;
        }
        break;
      case 'h':
        usage();
        exit(1);
        break;
      default :
        fprintf(stderr, "invalid switch: '-%c'\n", ch );
        usage();
        exit(1);
        break;
    }
  }

  if (errors) {
    fprintf(stderr, "%s: can't continue due to the above errors\n",
            progname );
    exit(1);
  }

  set_dbregs(dr7, w, 4, &d);

  fprintf(stdout, "\n" );
  fprintf(stdout, 
          "watch    status        type  len     address\n"
          "-----  --------  ----------  ---  ----------\n");
  for (i=0; i<4; i++) {
    if (d.dr7 & (0x03 << (i*2))) {
      unsigned type, len;
      type = (d.dr7 >> (16+(i*4))) & 3;
      len =  (d.dr7 >> (16+(i*4)+2)) & 3;
      fprintf(stdout, "%-5d  %-8s  %-10s  %3d  0x%08x\n",
              i, "enabled", watchtype_str(type), len+1, DBREG_DRX((&d),i));
    }
    else {
      fprintf(stdout, "%-5d  %-8s\n",
              i, w[i].active ? "enabled" : "disabled" );
    }
  }

  fprintf(stdout, "\n" );
  for (i=0; i<8; i++) {
    fprintf(stdout, "set $dr%i 0x%08x\n", i, DBREG_DRX((&d),i));
  }
  fprintf(stdout, "\n" );

  return 0;
}


